# the module includes adapted parts of the Python 2 and Python 3 standard
# libraries. It is intended to provide a clean implementation of the pickle
# protocol minus the actual file format on disk. This module contains no
# code specific to HDF5 files.
import sys
from collections import namedtuple
from pickle import (whichmodule, PicklingError, UnpicklingError,
                    dispatch_table, _extension_registry)


class H5itPicklingError(PicklingError):
    pass


class H5itUnpicklingError(UnpicklingError):
    pass


GlobalTuple = namedtuple(u'Global', (u'module', u'name'))


r_key_func = u'func'
r_key_cls = u'cls'
r_key_args = u'args'
r_key_state = u'state'
r_key_listitems = u'listitems'
r_key_dictitems = u'dictitems'


is_py2 = sys.version_info.major == 2
is_py3 = sys.version_info.major == 3


if is_py2:
    from types import TypeType, StringType, TupleType

if is_py3:
    from pickle import _getattribute, _compat_pickle


# adapted from Python 3 save_global
def save_global_py3(obj, name=None, proto=2, fix_imports=True):

    if name is None and proto >= 4:
        name = getattr(obj, '__qualname__', None)
    if name is None:
        name = obj.__name__

    module_name = whichmodule(obj, name, allow_qualname=proto >= 4)
    try:
        __import__(module_name, level=0)
        module = sys.modules[module_name]
        obj2 = _getattribute(module, name, allow_qualname=proto >= 4)
    except (ImportError, KeyError, AttributeError):
        raise PicklingError(
            "Can't pickle %r: it's not found as %s.%s" %
            (obj, module_name, name))
    else:
        if obj2 is not obj:
            raise PicklingError(
                "Can't pickle %r: it's not the same object as %s.%s" %
                (obj, module_name, name))

    if proto >= 2:
        code = _extension_registry.get((module_name, name))
        if code:
            # assert code > 0
            # if code <= 0xff:
            #     write(EXT1 + pack("<B", code))
            # elif code <= 0xffff:
            #     write(EXT2 + pack("<H", code))
            # else:
            #     write(EXT4 + pack("<i", code))
            # return
            raise H5itPicklingError("h5it Can't pickle %r: extension codes are not"
                            " supported yet." % obj)
    # Non-ASCII identifiers are supported only with protocols >= 3.
    if proto >= 4:
        # self.save(module_name)
        # self.save(name)
        # write(STACK_GLOBAL)
        raise H5itPicklingError("h5it Can't pickle %r: protocol %i is not "
                        "supported yet." % (obj, proto))
    elif proto >= 3:
        # write(GLOBAL + bytes(module_name, "utf-8") + b'\n' +
        #       bytes(name, "utf-8") + b'\n')
        raise H5itPicklingError("h5it Can't pickle %r: protocol %i is not "
                        "supported yet." % (obj, proto))
    else:
        if fix_imports:
            r_name_mapping = _compat_pickle.REVERSE_NAME_MAPPING
            r_import_mapping = _compat_pickle.REVERSE_IMPORT_MAPPING
            if (module_name, name) in r_name_mapping:
                module_name, name = r_name_mapping[(module_name, name)]
            if module_name in r_import_mapping:
                module_name = r_import_mapping[module_name]
        try:
            return GlobalTuple(bytes(module_name, "ascii"),
                               bytes(name, "ascii"))
        except UnicodeEncodeError:
            raise PicklingError(
                "can't pickle global identifier '%s.%s' using "
                "pickle protocol %i" % (module, name, proto))


# adapted from Python 2 save_global
def save_global_py2(obj, name=None, proto=2):

    if name is None:
        name = obj.__name__

    module = getattr(obj, "__module__", None)
    if module is None:
        module = whichmodule(obj, name)

    try:
        __import__(module)
        mod = sys.modules[module]
        klass = getattr(mod, name)
    except (ImportError, KeyError, AttributeError):
        raise PicklingError(
            "Can't pickle %r: it's not found as %s.%s" %
            (obj, module, name))
    else:
        if klass is not obj:
            raise PicklingError(
                "Can't pickle %r: it's not the same object as %s.%s" %
                (obj, module, name))

    if proto >= 2:
        code = _extension_registry.get((module, name))
        if code:
            # assert code > 0
            # if code <= 0xff:
            #     write(EXT1 + chr(code))
            # elif code <= 0xffff:
            #     write("%c%c%c" % (EXT2, code&0xff, code>>8))
            # else:
            #     write(EXT4 + pack("<i", code))
            # return
            raise H5itPicklingError(
                "h5it Can't pickle %r: extension codes "
                "are not supported yet." % obj)
    return GlobalTuple(module, name)


def save_py3(obj, proto=2):
    t = type(obj)

    # Check copyreg.dispatch_table only.
    reduce = dispatch_table.get(t)
    if reduce is not None:
        rv = reduce(obj)
    else:
        # Check for a class with a custom metaclass; treat as regular class
        try:
            issc = issubclass(t, type)
        except TypeError: # t is not a class (old Boost; see SF #502085)
            issc = False
        if issc:
            return save_global_py3(obj)

        # Check for a __reduce_ex__ method, fall back to __reduce__
        reduce = getattr(obj, "__reduce_ex__", None)
        if reduce is not None:
            rv = reduce(proto)
        else:
            reduce = getattr(obj, "__reduce__", None)
            if reduce is not None:
                rv = reduce()
            else:
                raise PicklingError("Can't pickle %r object: %r" %
                                    (t.__name__, obj))

    # Check for string returned by reduce(), meaning "save as global"
    if isinstance(rv, str):
        return save_global_py3(obj, rv)

    # Assert that reduce() returned a tuple
    if not isinstance(rv, tuple):
        raise PicklingError("%s must return string or tuple" % reduce)

    # Assert that it returned an appropriately sized tuple
    l = len(rv)
    if not (2 <= l <= 5):
        raise PicklingError("Tuple returned by %s must have "
                            "two to five elements" % reduce)

    # return the reduce() output for serializing
    return save_reduce_py3(obj=obj, *rv)


def save_py2(obj, proto=2):
    t = type(obj)

    # Check copy_reg.dispatch_table
    reduce = dispatch_table.get(t)
    if reduce:
        rv = reduce(obj)
    else:
        # Check for a class with a custom metaclass; treat as regular class
        try:
            issc = issubclass(t, TypeType)
        except TypeError: # t is not a class (old Boost; see SF #502085)
            issc = 0
        if issc:
            return save_global_py2(obj, proto=proto)

        # Check for a __reduce_ex__ method, fall back to __reduce__
        reduce = getattr(obj, "__reduce_ex__", None)
        if reduce:
            rv = reduce(proto)
        else:
            reduce = getattr(obj, "__reduce__", None)
            if reduce:
                rv = reduce()
            else:
                raise PicklingError("Can't pickle %r object: %r" %
                                    (t.__name__, obj))

    # Check for string returned by reduce(), meaning "save as global"
    if type(rv) is StringType:
        return save_global_py2(obj, rv)

    # Assert that reduce() returned a tuple
    if type(rv) is not TupleType:
        raise PicklingError("%s must return string or tuple" % reduce)

    # Assert that it returned an appropriately sized tuple
    l = len(rv)
    if not (2 <= l <= 5):
        raise PicklingError("Tuple returned by %s must have "
                            "two to five elements" % reduce)

    # Return the reduce() output for serialization
    return save_reduce_py2(obj=obj, *rv)


def save_reduce_py3(func, args, state=None, listitems=None, dictitems=None,
                    obj=None, proto=2):
    reduced = {}

    if not isinstance(args, tuple):
        raise PicklingError("args from save_reduce() must be a tuple")
    if not callable(func):
        raise PicklingError("func from save_reduce() must be callable")

    func_name = getattr(func, "__name__", "")
    if proto >= 4 and func_name == "__newobj_ex__":
        # cls, args, kwargs = args
        # if not hasattr(cls, "__new__"):
        #     raise PicklingError("args[0] from {} args has no __new__"
        #                         .format(func_name))
        # if obj is not None and cls is not obj.__class__:
        #     raise PicklingError("args[0] from {} args has the wrong class"
        #                         .format(func_name))
        # save(cls)
        # save(args)
        # save(kwargs)
        # write(NEWOBJ_EX)
        raise H5itPicklingError("h5it can't reduce {} {}: __newobj_ex__ is not "
                        "supported.".format(obj, func_name))
    elif proto >= 2 and func_name == "__newobj__":
        # A __reduce__ implementation can direct protocol 2 or newer to
        # use the more efficient NEWOBJ opcode, while still
        # allowing protocol 0 and 1 to work normally.  For this to
        # work, the function returned by __reduce__ should be
        # called __newobj__, and its first argument should be a
        # class.  The implementation for __newobj__
        # should be as follows, although pickle has no way to
        # verify this:
        #
        # def __newobj__(cls, *args):
        #     return cls.__new__(cls, *args)
        #
        # Protocols 0 and 1 will pickle a reference to __newobj__,
        # while protocol 2 (and above) will pickle a reference to
        # cls, the remaining args tuple, and the NEWOBJ code,
        # which calls cls.__new__(cls, *args) at unpickling time
        # (see load_newobj below).  If __reduce__ returns a
        # three-tuple, the state from the third tuple item will be
        # pickled regardless of the protocol, calling __setstate__
        # at unpickling time (see load_build below).
        #
        # Note that no standard __newobj__ implementation exists;
        # you have to provide your own.  This is to enforce
        # compatibility with Python 2.2 (pickles written using
        # protocol 0 or 1 in Python 2.3 should be unpicklable by
        # Python 2.2).
        cls = args[0]
        if not hasattr(cls, "__new__"):
            raise PicklingError(
                "args[0] from __newobj__ args has no __new__")
        if obj is not None and cls is not obj.__class__:
            raise PicklingError(
                "args[0] from __newobj__ args has the wrong class")
        args = args[1:]
        reduced[r_key_cls] = save_global_py3(cls)
        reduced[r_key_args] = args
    else:
        reduced[r_key_func] = save_global_py3(func)
        reduced[r_key_args] = args

    # if obj is not None:
    #     # If the object is already in the memo, this means it is
    #     # recursive. In this case, throw away everything we put on the
    #     # stack, and fetch the object back from the memo.
    #     if id(obj) in memo:
    #         write(POP + self.get(self.memo[id(obj)][0]))
    #     else:
    #         memoize(obj)

    # More new special cases (that work with older protocols as
    # well): when __reduce__ returns a tuple with 4 or 5 items,
    # the 4th and 5th item should be iterators that provide list
    # items and dict items (as (key, value) tuples), or None.

    if listitems is not None:
        reduced[r_key_listitems] = listitems

    if dictitems is not None:
        reduced[r_key_dictitems] = dictitems

    if state is not None:
        reduced[r_key_state] = state

    return reduced


def save_reduce_py2(func, args, state=None, listitems=None, dictitems=None,
                    obj=None, proto=2):
    reduced = {}
    # Assert that args is a tuple or None
    if not isinstance(args, TupleType):
        raise PicklingError("args from reduce() should be a tuple")

    # Assert that func is callable
    if not hasattr(func, '__call__'):
        raise PicklingError("func from reduce should be callable")

    # Protocol 2 special case: if func's name is __newobj__, use NEWOBJ
    if proto >= 2 and getattr(func, "__name__", "") == "__newobj__":
        # A __reduce__ implementation can direct protocol 2 to
        # use the more efficient NEWOBJ opcode, while still
        # allowing protocol 0 and 1 to work normally.  For this to
        # work, the function returned by __reduce__ should be
        # called __newobj__, and its first argument should be a
        # new-style class.  The implementation for __newobj__
        # should be as follows, although pickle has no way to
        # verify this:
        #
        # def __newobj__(cls, *args):
        #     return cls.__new__(cls, *args)
        #
        # Protocols 0 and 1 will pickle a reference to __newobj__,
        # while protocol 2 (and above) will pickle a reference to
        # cls, the remaining args tuple, and the NEWOBJ code,
        # which calls cls.__new__(cls, *args) at unpickling time
        # (see load_newobj below).  If __reduce__ returns a
        # three-tuple, the state from the third tuple item will be
        # pickled regardless of the protocol, calling __setstate__
        # at unpickling time (see load_build below).
        #
        # Note that no standard __newobj__ implementation exists;
        # you have to provide your own.  This is to enforce
        # compatibility with Python 2.2 (pickles written using
        # protocol 0 or 1 in Python 2.3 should be unpicklable by
        # Python 2.2).
        cls = args[0]
        if not hasattr(cls, "__new__"):
            raise PicklingError(
                "args[0] from __newobj__ args has no __new__")
        if obj is not None and cls is not obj.__class__:
            raise PicklingError(
                "args[0] from __newobj__ args has the wrong class")
        args = args[1:]
        reduced[r_key_cls] = save_global_py2(cls)
        reduced[r_key_args] = args
    else:
        reduced[r_key_func] = save_global_py2(func)
        reduced[r_key_args] = args

    # More new special cases (that work with older protocols as
    # well): when __reduce__ returns a tuple with 4 or 5 items,
    # the 4th and 5th item should be iterators that provide list
    # items and dict items (as (key, value) tuples), or None.

    if listitems is not None:
        reduced[r_key_listitems] = listitems

    if dictitems is not None:
        reduced[r_key_dictitems] = dictitems

    if state is not None:
        reduced[r_key_state] = state

    return reduced


def find_class_py3(module, name, proto=2, fix_imports=True):
    if proto < 3 and fix_imports:
        if (module, name) in _compat_pickle.NAME_MAPPING:
            module, name = _compat_pickle.NAME_MAPPING[(module, name)]
        if module in _compat_pickle.IMPORT_MAPPING:
            module = _compat_pickle.IMPORT_MAPPING[module]
    __import__(module, level=0)
    return _getattribute(sys.modules[module], name,
                         allow_qualname=proto >= 4)


def find_class_py2(module, name):
    # Subclasses may override this
    __import__(module)
    mod = sys.modules[module]
    klass = getattr(mod, name)
    return klass


def load_build_py2(inst, state):
    setstate = getattr(inst, "__setstate__", None)
    if setstate:
        setstate(state)
        return
    slotstate = None
    if isinstance(state, tuple) and len(state) == 2:
        state, slotstate = state
    if state:
        try:
            d = inst.__dict__
            try:
                for k, v in state.iteritems():
                    d[intern(k)] = v
            # keys in state don't have to be strings
            # don't blow up, but don't go out of our way
            except TypeError:
                d.update(state)

        except RuntimeError:
            # XXX In restricted execution, the instance's __dict__
            # is not accessible.  Use the old way of unpickling
            # the instance variables.  This is a semantic
            # difference when unpickling in restricted
            # vs. unrestricted modes.
            # Note, however, that cPickle has never tried to do the
            # .update() business, and always uses
            #     PyObject_SetItem(inst.__dict__, key, value) in a
            # loop over state.items().
            for k, v in state.items():
                setattr(inst, k, v)
    if slotstate:
        for k, v in slotstate.items():
            setattr(inst, k, v)
    return inst


def load_build_py3(inst, state):
    setstate = getattr(inst, "__setstate__", None)
    if setstate is not None:
        setstate(state)
        return
    slotstate = None
    if isinstance(state, tuple) and len(state) == 2:
        state, slotstate = state
    if state:
        inst_dict = inst.__dict__
        intern = sys.intern
        for k, v in state.items():
            if type(k) is str:
                inst_dict[intern(k)] = v
            else:
                inst_dict[k] = v
    if slotstate:
        for k, v in slotstate.items():
            setattr(inst, k, v)

    return inst


def load_global_py3(module, name):
    module = module.decode("utf-8")
    name = name.decode("utf-8")
    return find_class_py3(module, name)


def load_global_py2(module, name):
    return find_class_py2(module, name)


if is_py2:
    pickle_save_global = save_global_py2
    pickle_save = save_py2
    pickle_load_build = load_build_py2
    pickle_load_global = load_global_py2
elif is_py3:
    pickle_save_global = save_global_py3
    pickle_save = save_py3
    pickle_load_build = load_build_py3
    pickle_load_global = load_global_py3

############################### DISPATCH TABLES ###############################
#
# PYTHON 2 TYPE         PYTHON 3 TYPE   FUNCTION            NOTES
# -----------------------------------------------------------------------------
# NoneType              type(None)      save_none           -
# bool                  bool            save_bool           -
# IntType               -               save_int            -
# LongType              int             save_long           -
# FloatType             float           save_float          -
# StringType            bytes           save_{string,bytes} Jython (Py2 only)
# UnicodeType           str             save_{unicode,str}  -
# TupleType             tuple           save_tuple          recursive stuff
# ListType              list            save_list           -
# DictionaryType        dict            save_dict           -
# PyStringMap           PyStringMap     save_dict           Jython only
# InstanceType          -               save_instance       Old style class?
# -                     set             save_set            -
# -                     frozen_set      save_frozenset      -
# ClassType             -               save_global         -
# FunctionType          FunctionType    save_global         -
# BuiltinFunctionType   -               save_global         -
# TypeType              -               save_global         -
# -                     type            save_type           why not Py2?
###############################################################################


######################### RECONSTRUCTION OPCODES ##############################
#
# PYTHON 2  PYTHON 3    FUNCTION            NOTES
# -----------------------------------------------------------------------------
# NEWOBJ    NEWOBJ      load_newobj     -
# -         NEWOBJ_EX   load_newobj_ex  Protocol 4+
# REDUCE    REDUCE      load_reduce     -
# BUILD     BUILD       load_build      -
# INST      INST        load_inst       Old style instance
###############################################################################



from __future__ import unicode_literals
import tempfile

from h5it import dump, load


path = tempfile.mkstemp()[1]


class Foo(object):

    def __init__(self):
        self.a = 1
        self.b = 'a'
        self.c = {'x': None}
        self.d = None
        self.e = False
        self.f = (1, 5, 1)
        self.g = ['h', 142+32j, -159081.1340]

    def __eq__(self, other):
        return (self.a == other.a and
                self.b == other.b and
                self.c == other.c and
                self.d == other.d and
                self.e == other.e and
                self.f == other.f and
                self.g == other.g)


class FooCustom(Foo):

    def __getstate__(self):
        state = self.__dict__.copy()
        state['f'] = list(state['f'])
        state['e_another_name'] = not state.pop('e')
        return state

    def __setstate__(self, state):
        state['f'] = tuple(state['f'])
        state['e'] = not state.pop('e_another_name')
        self.__dict__.update(state)


def test_save_instance():
    dump(Foo(), path)


def test_load_instance():
    x = Foo()
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == Foo


def test_load_custom_instance():
    x = FooCustom()
    dump(x, path)
    y = load(path)
    assert y == x
    print(type(y))
    assert type(y) == FooCustom


def test_save_set():
    dump({'b', True, 'd', 1, None, ('key', 2.5012343)}, path)


def test_load_set():
    x = {'b', True, 'd', 1, None, ('key', 2.5012343)}
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == set



# coding=utf-8
import h5it
import pickle
import sys
import os

unicode_test_str = (u'σκουλικομερμυγκότρυπα ασπρη πέτρα '
                    u'ξέξασπρη κι από τον ήλιο ξεξασπρότερη')

byte_test_str = b'the rain in spain falls mainly on the plain'

v = sys.version_info.major

test_dir = os.path.dirname(os.path.realpath(__file__)) + '/'
print(test_dir)

with open(test_dir + 'py{}_unicode_proto2_bin.pickle'.format(v), 'wb') as f:
    pickle.dump(unicode_test_str, f, protocol=2)

with open(test_dir + 'py{}_bytes_proto2_bin.pickle'.format(v), 'wb') as f:
    pickle.dump(byte_test_str, f, protocol=2)

h5it.dump(unicode_test_str, test_dir + 'py{}_unicode.hdf5'.format(v))
h5it.dump(byte_test_str, test_dir + 'py{}_bytes.hdf5'.format(v))



# coding=utf-8
from __future__ import unicode_literals

import tempfile
from nose.tools import raises
import os
from os.path import join as j
import numpy as np
from pathlib import (Path, PosixPath, PurePosixPath,
                     WindowsPath, PureWindowsPath)

from h5it import dump, load, H5itUnpicklingError
from h5it.base import is_py2, host_is_posix, host_is_windows
import pickle

test_dir = os.path.dirname(os.path.realpath(__file__))
files_dir = os.path.join(test_dir, 'files')

unicode_test_str = (u'σκουλικομερμυγκότρυπα ασπρη πέτρα '
                    u'ξέξασπρη κι από τον ήλιο ξεξασπρότερη')

py2_unicode_pickle = j(files_dir, 'py2_unicode_proto2_bin.pickle')
py3_unicode_pickle = j(files_dir, 'py3_unicode_proto2_bin.pickle')
py2_unicode_h5it = j(files_dir, 'py2_unicode.hdf5')
py3_unicode_h5it = j(files_dir, 'py3_unicode.hdf5')

py2_bytes_pickle = j(files_dir, 'py2_bytes_proto2_bin.pickle')
py3_bytes_pickle = j(files_dir, 'py3_bytes_proto2_bin.pickle')
py2_bytes_h5it = j(files_dir, 'py2_bytes.hdf5')
py3_bytes_h5it = j(files_dir, 'py3_bytes.hdf5')

if is_py2:
    unicode_type = unicode
    bytes_type = str
else:
    unicode_type = str
    bytes_type = bytes

path = tempfile.mkstemp()[1]


def test_save_integer():
    dump(1, path)


def test_save_float():
    dump(-14.1512, path)


def test_save_complex():
    dump(25.1512 + 7j, path)


def test_save_unicode():
    dump(unicode_test_str, path)


def test_save_byte_str():
    dump(b'byte str', path)


def test_save_bool():
    dump(True, path)


def test_save_none():
    dump(None, path)


def test_save_ndarray():
    x = np.random.rand(151, 16, 16, 1)
    dump(x, path)


def test_save_path():
    dump(Path('/some/path/here'), path)


def test_save_empty_list():
    dump([], path)


def test_save_list():
    dump([1, 'a', None, True], path)


def test_save_recursive_list():
    dump([1, 'a', None, True, [b'another', -125.14]], path)


def test_save_empty_tuple():
    dump(tuple(), path)


def test_save_tuple():
    dump((1, 'xyx', 15.161), path)


def test_save_recursive_tuple():
    dump((1, 'xyx', 15.161, (None, {'a': 1}, [1, 3, 4], True)), path)


def test_save_empty_dict():
    dump({}, path)


def test_save_dict_with_non_string_keys():
    dump({1: 'a', None: True}, path)


def test_save_dict():
    dump({'b': 2, 'c': True}, path)


def test_save_dict_recursive():
    dump({'b': 2, 'c': True, 'd': [1, None, {'key': 2.5012343}]}, path)


def test_load_integer():
    x = 1
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == int


def test_load_float():
    x = +11241.151214541
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == float


def test_load_complex():
    x = -1589.151214541 - 1390815.24155j
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == complex


def test_load_unicode():
    dump(unicode_test_str, path)
    y = load(path)
    assert y == unicode_test_str
    assert type(y) == unicode_type


def test_load_byte_str():
    x = b"some byte str"
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == bytes_type


def test_load_bool():
    x = False
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == bool


def test_load_none():
    x = None
    dump(x, path)
    y = load(path)
    assert y == x
    assert isinstance(y, type(None))
    assert y is None


def test_load_ndarray():
    x = np.random.rand(151, 16, 16, 1)
    dump(x, path)
    y = load(path)
    assert np.all(y == x)
    assert type(y) == np.ndarray


if host_is_posix:
    def test_load_posix_path_on_posix():
        x = PosixPath('/some/path/here')
        dump(x, path)
        y = load(path)
        assert y == x
        assert type(y) == PosixPath

    def test_load_windows_path_on_posix():
        x = PureWindowsPath('C:\some\path\here')
        dump(x, path)
        y = load(path)
        assert y == x
        assert type(y) == PureWindowsPath


if host_is_windows:
    def test_load_posix_path_on_windows():
        x = PurePosixPath('/some/path/here')
        dump(x, path)
        y = load(path)
        assert y == x
        assert type(y) == PurePosixPath

    def test_load_windows_path_on_windows():
        x = WindowsPath('C:\some\path\here')
        dump(x, path)
        y = load(path)
        assert y == x
        assert type(y) == WindowsPath


def test_load_empty_list():
    x = []
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == list


def test_load_list():
    x = [1, 'a', None, True]
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == list


def test_load_recursive_list():
    x = [1, 'a', None, True, ['another', -125.14]]
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == list


def test_load_empty_tuple():
    x = tuple()
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == tuple


def test_load_tuple():
    x = (1, 'xyx', 15.161)
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == tuple


def test_load_recursive_tuple():
    x = (1, 'xyx', 15.161, (None, {'a': 1}, [1, 3, 4], True))
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == tuple


def test_load_empty_dict():
    x = {}
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == dict


def test_load_dict():
    x = {'b': 2, 'c': True}
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == dict


def test_load_dict_with_non_string_keys():
    x = {1: 'a', None: True}
    dump(x, path)
    y = load(path)
    assert y == x


def test_load_recursive_dict():
    x = {'b': 2, 'c': True, 'd': [1, None, {'key': 2.5012343}]}
    dump(x, path)
    y = load(path)
    assert y == x
    assert type(y) == dict


def test_load_reference():
    c = [1, 2, 3]
    a = {'c_from_a': c}
    b = {'c_from_b': c}
    dump((a, b), path)
    a_l, b_l = load(path)
    assert id(a_l['c_from_a']) == id(b_l['c_from_b'])


def test_save_with_path():
    dump(None, Path(path))


def test_save_with_path_load_with_str():
    dump(None, Path(path))
    assert load(path) is None


def test_save_with_path_load_with_path():
    dump(None, Path(path))
    assert load(Path(path)) is None


def test_save_with_str_load_with_path():
    dump(None, path)
    assert load(Path(path)) is None


if is_py2:

    def test_load_unicode_from_py3_on_py2():
        with open(py3_unicode_pickle, 'rb') as f:
            x = pickle.load(f)
        y = load(py3_unicode_h5it)
        assert type(y) == type(x)
        assert x == y

    def test_load_bytes_from_py3_on_py2():
        with open(py3_bytes_pickle, 'rb') as f:
            x = pickle.load(f)
        y = load(py3_bytes_h5it)
        assert type(y) == type(x)
        assert x == y

    def test_load_unicode_from_py2_on_py2():
        with open(py2_unicode_pickle, 'rb') as f:
            x = pickle.load(f)
        y = load(py2_unicode_h5it)
        assert type(y) == type(x)
        assert x == y

    def test_load_bytes_from_py2_on_py2():
        with open(py2_bytes_pickle, 'rb') as f:
            x = pickle.load(f)
        y = load(py2_bytes_h5it)
        assert type(y) == type(x)
        assert x == y

else:

    def test_load_unicode_from_py3_on_py3():
        with open(py3_unicode_pickle, 'rb') as f:
            x = pickle.load(f)
        y = load(py3_unicode_h5it)
        assert type(y) == type(x)
        assert x == y

    def test_load_bytes_from_py3_on_py3():
        with open(py3_bytes_pickle, 'rb') as f:
            x = pickle.load(f)
        y = load(py3_bytes_h5it)
        assert type(y) == type(x)
        assert x == y

    def test_load_bytes_from_py3_on_py3_encoding_ascii_shouldnt_matter():
        with open(py3_bytes_pickle, 'rb') as f:
            x = pickle.load(f)
        y = load(py3_bytes_h5it, encoding='ASCII')
        assert type(y) == type(x)
        assert x == y

    def test_load_unicode_from_py2_on_py3():
        with open(py2_unicode_pickle, 'rb') as f:
            x = pickle.load(f)
        y = load(py2_unicode_h5it)
        assert type(y) == type(x)
        assert x == y

    def test_load_bytes_from_py2_on_py3_encoding_default():
        with open(py2_bytes_pickle, 'rb') as f:
            x = pickle.load(f)
        y = load(py2_bytes_h5it)
        assert type(y) == type(x)
        assert x == y

    def test_load_bytes_from_py2_on_py3_encoding_ascii():
        with open(py2_bytes_pickle, 'rb') as f:
            x = pickle.load(f, encoding='ASCII')
        y = load(py2_bytes_h5it, encoding='ASCII')
        assert type(y) == type(x)
        assert x == y

    def test_load_bytes_from_py2_on_py3_encoding_bytes():
        with open(py2_bytes_pickle, 'rb') as f:
            x = pickle.load(f, encoding='bytes')
        y = load(py2_bytes_h5it, encoding='bytes')
        assert type(y) == type(x)
        assert x == y

    @raises(H5itUnpicklingError)
    def test_load_invalid_encoding():
        load(py2_bytes_pickle, encoding='asda')



from .base import load, dump  # main API for saving and loading files.
from .stdpickle import H5itPicklingError, H5itUnpicklingError

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions




# This file was generated by 'versioneer.py' (0.11) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '0.5.0-2-g1889870'
version_full = '18898707b721ede2a3d3e8ba5ed7fbd22e409389'
def get_versions(default={}, verbose=False):
    return {'version': version_version, 'full': version_full}




from __future__ import unicode_literals

import os
from collections import namedtuple
from pathlib import PosixPath, WindowsPath, PurePosixPath, PureWindowsPath
import numpy as np
import h5py

from .stdpickle import (H5itPicklingError, H5itUnpicklingError,
                        is_py2, is_py3,
                        r_key_func, r_key_cls, r_key_args,
                        r_key_state, r_key_listitems, r_key_dictitems,
                        pickle_load_global, pickle_save_global, GlobalTuple,
                        pickle_load_build, pickle_save)

if is_py2:
    from types import ClassType, FunctionType, BuiltinFunctionType, TypeType
    strType = unicode
    py2_bytesType = str
    py3_bytesType = tuple()  # will never encounter Py3 bytes str on Py2
    as_unicode_str = unicode

    numberTypes = int, long, float, complex
    globalTypes = ClassType, FunctionType, BuiltinFunctionType, TypeType
elif is_py3:
    from types import FunctionType
    strType = str
    py2_bytesType = tuple()  # will never encounter Py2 bytes str on Py3
    py3_bytesType = bytes
    as_unicode_str = str

    numberTypes = int, float, complex
    globalTypes = FunctionType
else:
    raise Exception('h5it is only compatible with Python 2 or Python 3')

host_is_posix = os.name == 'posix'
host_is_windows = os.name == 'nt'

attr_key_type = 'type'
attr_key_type_reduction = 'reduction'

attr_key_reduction_cls_module = 'cls_module'
attr_key_reduction_cls_name = 'cls_name'
attr_key_reduction_func_module = 'func_module'
attr_key_reduction_func_name = 'func_name'

attr_key_global_module = 'module'
attr_key_global_name = 'name'

attr_key_number_value = 'number_value'
attr_key_bool_value = 'bool_value'

top_level_group_namespace = 'h5it'


# ------------------------------ IMPORTS ------------------------------ #

ListItem = namedtuple('ListItem', ['i', 'item'])


def load_list(parent, name, memo, encoding):
    node = parent[name]
    unordered = []
    for i, j in enumerate(node.keys()):
        unordered.append(ListItem(int(j), h5_import(node, j, memo, encoding)))
    ordered = sorted(unordered)
    counts = [x.i for x in ordered]
    items = [x.item for x in ordered]
    if counts != list(range(len(counts))):
        raise H5itUnpicklingError("Attempted to import a list "
                                  "that is missing elements")
    return items


def load_tuple(parent, name, memo, encoding):
    return tuple(load_list(parent, name, memo, encoding))


def load_unicode_dict(parent, name, memo, encoding):
    node = parent[name]
    imported_dict = {}
    for k in node.keys():
        imported_dict[k] = h5_import(node, k, memo, encoding)
    return imported_dict


def load_dict(parent, name, memo, encoding):
    node = parent[name]
    return dict(h5_import(node, k, memo, encoding) for k in node.keys())


def load_reducible(parent, name, memo, encoding):
    node = parent[name]
    # import the class and new it up
    if attr_key_global_name in node.attrs:
        # reduction actually saved out a global
        # node that it is safe to skip h5_import and hence the memo checking
        # here as we are calling load_global on the exact same node that
        # load_reducible was called on. Prior to invocation a check was made of
        # node to ensure it wasn't a symlink so we are OK.
        return load_global(parent, name, memo, encoding)

    # if not, we are loading with NEWOBJ or REDUCE
    args = load_tuple(node, r_key_args, memo, encoding)
    if attr_key_reduction_cls_module in node.attrs:
        # reduction using the NEWOBJ protocol
        cls_module = node.attrs[attr_key_reduction_cls_module]
        cls_name = node.attrs[attr_key_reduction_cls_name]
        cls = pickle_load_global(cls_module, cls_name)
        obj = cls.__new__(cls, *args)
    elif attr_key_reduction_func_module in node.attrs:
        # reduction using the REDUCE protocol
        func_module = node.attrs[attr_key_reduction_func_module]
        func_name = node.attrs[attr_key_reduction_func_name]
        func = pickle_load_global(func_module, func_name)
        obj = func(*args)
    else:
        raise H5itUnpicklingError(
            "error loading reduction - can't find {} or {} in attrs".format(
            attr_key_reduction_cls_module, attr_key_reduction_func_module))

    if r_key_state in node:
        state = h5_import(node, r_key_state, memo, encoding)
        pickle_load_build(obj, state)

    if r_key_listitems in node:
        listitems = h5_import(node, r_key_listitems, memo, encoding)
        for i in listitems:
            obj.append(i)

    if r_key_dictitems in node:
        iteritems = h5_import(node, r_key_dictitems, memo, encoding)
        for k, v in iteritems:
            obj[k] = v

    return obj


def load_global(parent, name, memo, encoding):
    module = parent[name].attrs[attr_key_global_module]
    m_name = parent[name].attrs[attr_key_global_name]
    return pickle_load_global(module, m_name)


def load_ndarray(parent, name, memo, encoding):
    return parent[name].value


def load_none(parent, name, memo, encoding):
    return None


def load_str(parent, name, memo, encoding):
    return parent[name].value


def load_bytes(parent, name, memo, encoding):
    return parent[name].value.tobytes()


def load_py2_bytes_on_py3(parent, name, memo, encoding):
    if encoding == 'ASCII':
        return parent[name].value.tobytes().decode('ASCII')
    elif encoding == 'bytes':
        return parent[name].value.tobytes()
    else:
        raise H5itUnpicklingError("The only valid encodings are 'ASCII' or "
                                  "'bytes'")

if is_py3:
    load_py2_bytes = load_py2_bytes_on_py3
else:
    load_py2_bytes = load_bytes


def load_bool(parent, name, memo, encoding):
    return bool(parent[name].attrs[attr_key_bool_value])


def load_number(parent, name, memo, encoding):
    return np.asscalar(parent[name].attrs[attr_key_number_value])


def load_posix_path(parent, name, memo, encoding):
    str_path = load_str(parent, name, memo, encoding)
    if host_is_posix:
        return PosixPath(str_path)
    else:
        return PurePosixPath(str_path)


def load_windows_path(parent, name, memo, encoding):
    str_path = load_str(parent, name, memo, encoding)
    if host_is_windows:
        return WindowsPath(str_path)
    else:
        return PureWindowsPath(str_path)


# ------------------------------ EXPORTS ------------------------------ #

zero_padded = lambda x: "{:0" + as_unicode_str(len(as_unicode_str(x))) + "}"


def save_list(l, parent, name, memo):
    list_node = parent.create_group(name)
    padded = zero_padded(len(l))
    for i, x in enumerate(l):
        h5_export(x, list_node, padded.format(i), memo)


is_string_keyed_dict = lambda d: (sum(not isinstance(k, strType)
                                      for k in d.keys()) == 0)


def save_unicode_dict(d, parent, name, memo):
    dict_node = parent.create_group(name)
    if not is_string_keyed_dict(d):
        raise ValueError("Only dictionaries with string keys can be "
                         "serialized")
    for k, v in d.items():
        h5_export(v, dict_node, str(k), memo)


def save_dict(d, parent, name, memo):
    dict_node = parent.create_group(name)
    for k, v in d.items():
        h5_export((k, v), dict_node, str(hash(k)), memo)


def save_reducible(x, parent, name, memo):
    # save down the object: we'll either get back a global or a reduction state
    reduction = pickle_save(x)
    if type(reduction) == GlobalTuple:
        save_global(reduction, parent, name, memo)
        return

    # Reduction is a dict that is ready to directly be saved. Let's make a
    # group
    node = parent.create_group(name)

    if r_key_cls in reduction:
        cls_module, cls_name = reduction[r_key_cls]
        node.attrs[attr_key_reduction_cls_module] = cls_module
        node.attrs[attr_key_reduction_cls_name] = cls_name
    elif r_key_func in reduction:
        func_module, func_name = reduction[r_key_func]
        node.attrs[attr_key_reduction_func_module] = func_module
        node.attrs[attr_key_reduction_func_name] = func_name
    else:
        H5itPicklingError("reduction state is missing a 'func' or 'cls'")

    # save out the reduction args
    save_list(reduction[r_key_args], node, r_key_args, memo)

    if r_key_state in reduction:
        h5_export(reduction[r_key_state], node, r_key_state, memo)

    if r_key_listitems in reduction:
        h5_export(reduction[r_key_listitems], node, r_key_listitems, memo)

    if r_key_dictitems in reduction:
        h5_export(reduction[r_key_dictitems], node, r_key_dictitems, memo)


def save_global(g, parent, name, _):
    node = parent.create_group(name)  # A blank group
    module, g_name = pickle_save_global(g)
    node.attrs[attr_key_global_module] = module
    node.attrs[attr_key_global_name] = g_name


def save_ndarray(a, parent, name, _):
    # fletcher32 is a checksum, gzip compression is supported by Matlab
    parent.create_dataset(name, data=a, compression='gzip', fletcher32=True)


def save_none(none, parent, name, _):
    parent.create_group(name)  # A blank group


def save_str(s, parent, name, _):
    dt = h5py.special_dtype(vlen=as_unicode_str)
    parent.create_dataset(name, data=s, dtype=dt)


def save_bytes(s, parent, name, _):
    parent.create_dataset(name, data=np.void(s))


def save_bool(a_bool, parent, name, _):
    group = parent.create_group(name)  # A blank group
    group.attrs[attr_key_bool_value] = a_bool


def save_number(a_number, parent, name, _):
    group = parent.create_group(name)  # A blank group
    group.attrs[attr_key_number_value] = a_number


def save_path(path, parent, name, _):
    parent.create_dataset(name, data=as_unicode_str(path))


T = namedtuple('T', ["type", "str", "importer", "exporter"])


types = [T(list, "list", load_list, save_list),
         T(tuple, "tuple", load_tuple, save_list),  # export is as list
         T(dict, "dict", load_dict, save_dict),
         T(np.ndarray, "ndarray", load_ndarray, save_ndarray),
         T(type(None), "NoneType", load_none, save_none),
         T(strType, "str", load_str, save_str),
         T(py2_bytesType, "py2_bytes", load_py2_bytes, save_bytes),
         T(py3_bytesType, "bytes", load_bytes, save_bytes),
         T(bool, "bool", load_bool, save_bool),
         T(globalTypes, "global", load_global, save_global),
         T(numberTypes, "Number", load_number, save_number),
         T((PosixPath, PurePosixPath), "pathlib.PosixPath",
           load_posix_path, save_path),
         T((WindowsPath, PureWindowsPath), "pathlib.WindowsPath",
           load_windows_path, save_path)]

type_to_exporter = dict()
str_to_importer = dict()
type_to_str = dict()

for t in types:
    # tuples of types are allowed, just add each type in turn with the same
    # rule
    if isinstance(t.type, tuple):
        for t_i in t.type:
            type_to_exporter[t_i] = t.exporter
            type_to_str[t_i] = t.str
    else:
        type_to_exporter[t.type] = t.exporter
        type_to_str[t.type] = t.str
    str_to_importer[t.str] = t.importer
    type_to_str[t.type] = t.str

# add on the reduction importer
str_to_importer[attr_key_type_reduction] = load_reducible


def link_path_if_softlink(node, name):
    if node.get(name, getclass=True, getlink=True) == h5py.SoftLink:
        # this node is a softlink - grab it's path
        return node.get(name, getlink=True).path


def h5_import(parent, name, memo, encoding):
    link_path = link_path_if_softlink(parent, name)
    if link_path is not None:
        # this node is a softlink - memoize the link destination path
        memo_path = link_path
    else:
        memo_path = parent[name].name
    if memo_path in memo:
        # this object has already been loaded - just return it
        return memo[memo_path]
    node = parent[name]
    type_ = node.attrs.get(attr_key_type)
    if type_ is not None:
        # node type is specific
        importer = str_to_importer.get(type_)
        if importer is not None:
            obj = importer(parent, name, memo, encoding)
            # remember we imported this already
            memo[memo_path] = obj
            return obj
        else:
            raise H5itUnpicklingError(
                "Don't know how to import type "
                "{} for node {}".format(type_, node))
    else:
        raise H5itUnpicklingError("Cannot find {} "
                                  "attribute on {}".format(attr_key_type,
                                                           node))


def h5_export(x, parent, name, memo):
    if id(x) in memo:
        # this object is already exported, just softlink to it.
        parent[name] = h5py.SoftLink(memo[id(x)].name)
        return
    type_x = type(x)
    exporter = type_to_exporter.get(type_x)
    if exporter is None:
        # use the pickle protocol to save a reducible/global
        exporter = save_reducible
        type_str = attr_key_type_reduction
    else:
        type_str = type_to_str.get(type_x)
    # definitely have type_str and exporter
    exporter(x, parent, name, memo)
    new_node = parent[name]
    new_node.attrs[attr_key_type] = type_str
    # remember we have exported this object
    memo[id(x)] = new_node
    # make sure that object doesn't die, as otherwise future objects
    # might reuse the same id. We steal the standard lib approach, and
    # just append on a special list in the memo (the list is stored on
    # the hash of the memo, so we are pretty guaranteed that no one
    # will use it!)
    try:
        memo[id(memo)].append(x)
    except KeyError:
        # I'm the first!
        memo[id(memo)] = [x]


def norm_path(path):
    r"""
    Uses all the tricks in the book to expand a path out to an absolute one.
    """
    return os.path.abspath(os.path.normpath(
        os.path.expandvars(os.path.expanduser(as_unicode_str(path)))))


def dump(x, path):
    with h5py.File(norm_path(path), "w") as f:
        h5_export(x, f, top_level_group_namespace, {})


def load_py2(path):
    with h5py.File(norm_path(path), "r") as f:
        # encoding is not used on Python 2, set to a dummy string
        return h5_import(f, top_level_group_namespace, {}, '')


def load_py3(path, encoding='ASCII'):
    if encoding not in ['ASCII', 'bytes']:
        raise H5itUnpicklingError("The only valid encodings are 'ASCII' or "
                                  "'bytes'")
    with h5py.File(norm_path(path), "r") as f:
        return h5_import(f, top_level_group_namespace, {}, encoding)

if is_py3:
    load = load_py3
else:
    load = load_py2
